""" Module containing Rink, NHLRink, NWHLRink, and IIHFRink for drawing and plotting ice surfaces. """


from hockey_rink._rink_plot import BaseRinkPlot
import hockey_rink.rink_feature as rf
from matplotlib.transforms import Affine2D
import numpy as np


__all__ = ["Rink", "NHLRink", "NWHLRink", "IIHFRink"]


class Rink(BaseRinkPlot):
    """ Rink to draw and plot on with matplotlib.

    Allows for customization to support any number of different possible rink dimensions.

    Default coordinate system and feature dimensions correspond to those used by the NHL.
        ie) All lengths are measured in feet.
            x-coordinates go from -100 to 100 and y-coordinates from -42.5 to 42.5.

    Attributes:
        rotation: matplotlib Affine2D
            Rotation of rink features when drawing.

        x_shift: float
            Amount x-coordinates are to be shifted.

            When viewing the rink horizontally, the coordinate of the center of the ice surface from left to right.
                eg) If using data with a coordinate system that goes from 0 to 200, x_shift should be 100.

            The actual coordinates won't be affected.  The purpose is to update the coordinates passed in to
            align with the drawing, not to alter the drawing to align with the coordinates.

        y_shift: float
            Amount y-coordinates are to be shifted.

            When viewing the rink horizontally, the coordinate of the center of the ice surface from bottom to top.
                eg) If using data with a coordinate system that goes from 0 to 85, y_shift should be 42.5.

            The actual coordinates won't be affected.  The purpose is to update the coordinates passed in to
            align with the drawing, not to alter the drawing to align with the coordinates.
    """

    def __init__(self, rotation=0, x_shift=0, y_shift=0,
                 line_thickness=1 / 6, line_color="red", line_zorder=5,
                 x_dot_to_lines=2, y_dot_to_lines=9 / 12,
                 boards=None, nzone=None, ozone=None, dzone=None,
                 red_line=None, blue_line=None, goal_line=None,
                 trapezoid=None, ref_circle=None, center_dot=None,
                 center_circle=None, faceoff_circle=None, faceoff_dot=None,
                 faceoff_lines=None, crease=None, crease_outline=None,
                 crossbar=None, net=None, **added_features):
        """ Initialize and create the features of the rink.

        Most parameters expect a dict with key/value pairs corresponding to RinkFeature attributes as well as one
        key for indicating the type of RinkFeature class being used.
            eg)
                feature_name = {
                    "class": feature_class,
                    "x": feature_x,
                    "y": feature_y
                    "length": feature_length,
                    "width": feature_width,
                    "thickness": feature_thickness,
                    "radius": feature_radius,
                    "resolution": feature_resolution,
                    "reflect_x": feature_reflect_x,
                    "reflect_y": feature_reflect_y
                    "is_constrained": feature_is_constrained,
                    "visible": feature_visible,
                    ...
                }
        Explanations for the attributes can be found in the RinkFeature documentation.

        The exceptions are:
            Multiple x and y coordinates can be passed as an array_like value.  If multiple values are provided,
            one feature will be created for each combination of coordinates.

            x and y values for faceoff lines correspond to the nearest faceoff dot.  Each coordinate will be
            included in four L shapes (above and right, below and left, etc) with the shape being altered
            accordingly.  The exact coordinate is determined by the values passed to x_dot_to_lines and y_dot_to_lines.

        Other attributes can be provided so long as they can be used by matplotlib's Polygon (such as color or zorder)
        or are appropriate for that particular feature (eg CircularImage requires a path attribute).

        All parameters that expect a dict only require any desired changes to be included in the key/value pairs.
            ie) To update the length of the boards (and, thereby, the rink), all that needs to be passed in is:
                boards={"length": new_length}
        Including other attributes in the dict is unnecessary unless they too require updates.

        Any parameters not included default to NHL dimensions, though they may be affected by changes to other
        parameters.

        To remove a feature from the rink, set visible to False.
            eg) trapezoid={"visible": False}

        New features can be included by passing in a dict with a name not included in the parameter list.
            eg) new_feature: {"class": feature_class, ...}

        The default zorders are:
            1: nzone, ozone, dzone, crease
            5: goal_line, trapezoid, ref_circle, center_circle, faceoff_circle,
                faceoff_dot, faceoff_lines, crease_outline, net
            6: crossbar
            10: red_line, blue_line
            100: boards

        Parameters:
            x_shift: float; default: 0
                Amount x-coordinates are to be shifted.

                When viewing the rink horizontally, the coordinate of the center of the ice surface from left to right.
                    eg) If using data with a coordinate system that goes from 0 to 200, x_shift should be 100.

                The actual coordinates won't be affected.  The purpose is to update the coordinates passed in to
                align with the drawing, not to alter the drawing to align with the coordinates.

            y_shift: float; default: 0
                Amount y-coordinates are to be shifted.

                When viewing the rink horizontally, the coordinate of the center of the ice surface from bottom to top.
                    eg) If using data with a coordinate system that goes from 0 to 85, y_shift should be 42.5.

                The actual coordinates won't be affected.  The purpose is to update the coordinates passed in to
                align with the drawing, not to alter the drawing to align with the coordinates.

            line_thickness: float; default: 1/6
                Thickness of all the thin lines on the ice (eg the goal line and faceoff circles) if not
                otherwise updated.

            line_color: color; default: "red"
                Color of all the thin lines on the ice (eg the goal line and faceoff circles) if not
                otherwise updated.

                An example of how to specify colors can be found at the following link:
                    https://matplotlib.org/stable/tutorials/colors/colors.html#sphx-glr-tutorials-colors-colors-py

            line_zorder: float; default: 5
                The zorder of all thin lines on the ice (eg the goal line and faceoff circles) if not
                otherwise updated.

                Determines which features are drawn first (lower values will cause features to appear under
                other features they may overlap).

            x_dot_to_lines: float; default: 2
                Length-wise distance between a faceoff dot and the L shapes in the faceoff circle.

            y_dot_to_lines = float; default: 9/12
                Width-wise distance between a faceoff dot and the L shapes in the faceoff circle.

            boards: dict; optional
                Attributes to update for the boards.

                Also affects the constraint that prevents features from extending outside the boards.

                The following won't be updated even if values are provided:
                    class
                    x
                    y
                    reflect_x
                    reflect_y
                    is_constrained

            nzone: dict; optional
                Attributes to update for the neutral zone.

                The following won't be updated even if values are provided:
                    class
                    x
                    y
                    width
                    reflect_x
                    reflect_y

            ozone: dict; optional
                Attributes to update for the offensive zone (largest x values).

                The following won't be updated even if values are provided:
                    class
                    x
                    y
                    length
                    reflect_x
                    reflect_y

            dzone: dict; optional
                Attributes to update for the defensive zone (smallest x values).

                The following won't be updated even if values are provided:
                    class
                    x
                    y
                    length
                    reflect_x
                    reflect_y

            red_line: dict; optional
                Attributes to update for the red (center) line.

            blue_line: dict; optional
                Attributes to update for the blue line(s).

            goal_line: dict; optional
                Attributes to update for the goal line(s).

            trapezoid: dict; optional
                Attributes to update for the trapezoid(s) for a goalie's restricted area.

            ref_circle: dict; optional
                Attributes to update for the ref's half circle.

            center_dot: dict; optional
                Attributes to update for the center faceoff dot.

            center_circle: dict; optional
                Attributes to update for the center faceoff circle.

            faceoff_circle: dict; optional
                Attributes to update for the faceoff circle(s) not including the one at
                center ice unless otherwise updated.

                All circles will include hashmarks by default.  To remove hashmarks, change
                the class to RinkCircle (ie faceoff_circle={"class": RinkCircle}).

            faceoff_dot: dict; optional
                Attributes to update for the faceoff dot(s) not including the one at
                center ice unless otherwise updated.

            faceoff_lines: dict; optional
                Attributes to update for the L shapes in the faceoff circles.

            crease: dict; optional
                Attributes to update for the crease(s).

            crease_outline: dict; optional
                Attributes to update for the outline(s) of the crease.

            crossbar: dict; optional
                Attributes to update for the crossbar(s) of the net.

            net: dict; optional
                Attributes to update for the netting of the net.

            added_features: dict; optional
                Any additional features to be added to the rink along with a dict of their
                attributes to be used.
        """

        self._rotation = Affine2D().rotate_deg(rotation)

        self.x_shift = x_shift
        self.y_shift = y_shift

        self._features = []
        self._feature_xlim = None
        self._feature_ylim = None

        boards = boards or {}
        board_params = {
            "length": 200,
            "width": 85,
            "thickness": 1,
            "radius": 28,
            "color": "black",
            "zorder": 100,
        }
        required_boards = {
            "class": rf.Boards,
            "x": 0,
            "y": 0,
            "reflect_x": True,
            "reflect_y": True,
            "is_constrained": False,
        }
        board_params = {**board_params, **boards, **required_boards}
        self._initialize_feature(board_params)

        boards_constraint = {
            "class": rf.BoardsConstraint,
            "length": board_params["length"],
            "width": board_params["width"],
            "radius": board_params["radius"],
            "thickness": board_params["thickness"],
            "reflect_x": False,
            "reflect_y": False,
            "visible": False,
        }
        self._initialize_feature(boards_constraint)
        self._boards_constraint = self._features.pop(-1)

        half_length = board_params["length"] / 2
        half_width = board_params["width"] / 2

        nzone = nzone or {}
        nzone_params = {
            "length": 50,
            "visible": False,
        }
        required_nzone = {
            "class": rf.RinkRectangle,
            "x": 0,
            "y": 0,
            "width": board_params["width"],
            "reflect_x": False,
            "reflect_y": False,
        }
        nzone_params = {**nzone_params, **nzone, **required_nzone}
        self._initialize_feature(nzone_params)

        self._half_nzone_length = nzone_params["length"] / 2

        ozone = ozone or {}
        ozone_length = half_length - self._half_nzone_length
        ozone_params = {
            "visible": False,
        }
        required_ozone = {
            "class": rf.RinkRectangle,
            "x": ozone_length / 2 + self._half_nzone_length,
            "y": 0,
            "length": ozone_length,
            "width": board_params["width"],
            "reflect_x": False,
            "reflect_y": False,
        }
        ozone_params = {**ozone_params, **ozone, **required_ozone}
        self._initialize_feature(ozone_params)

        dzone = dzone or {}
        dzone_params = {
            "visible": False,
        }
        required_dzone = {
            "class": rf.RinkRectangle,
            "x": -ozone_params["x"],
            "y": 0,
            "length": ozone_length,
            "width": board_params["width"],
            "reflect_x": False,
            "reflect_y": False,
        }
        dzone_params = {**dzone_params, **dzone, **required_dzone}
        self._initialize_feature(dzone_params)

        red_line = red_line or {}
        red_line_params = {
            "class": rf.RinkRectangle,
            "length": 1,
            "width": board_params["width"],
            "color": "red",
            "zorder": 10,
        }
        red_line_params.update(red_line)
        self._initialize_feature(red_line_params)

        blue_line = blue_line or {}
        blue_line_params = {
            "class": rf.RinkRectangle,
            "length": 1,
            "width": board_params["width"],
            "reflect_x": True,
            "color": "blue",
            "zorder": 10,
        }
        blue_line_params.update(blue_line)
        blue_line_params["x"] = blue_line_params.get(
            "x", self._half_nzone_length + blue_line_params["length"] / 2)

        self._blue_line_thickness = blue_line_params["length"]
        self._initialize_feature(blue_line_params)

        goal_line = goal_line or {}
        goal_line_params = {
            "class": rf.RinkRectangle,
            "length": line_thickness,
            "width": board_params["width"],
            "reflect_x": True,
            "color": line_color,
            "zorder": line_zorder,
        }
        goal_line_params.update(goal_line)

        # back edge of goal line is 11' from the boards
        goal_line_params["x"] = goal_line_params.get(
            "x", half_length - 11 - goal_line_params["length"] / 2)

        self._initialize_feature(goal_line_params)

        # trapezoid lines go from 11 ft from center to 14 ft from center
        # 11 and 14 refer to the center of the line
        trapezoid = trapezoid or {}
        trapezoid_params = {
            "class": rf.TrapezoidLine,
            "x": goal_line_params["x"] + goal_line_params["length"] / 2,
            "y": 11,
            "width": 3,
            "thickness": line_thickness,
            "reflect_x": True,
            "reflect_y": True,
            "color": line_color,
            "zorder": line_zorder,
        }
        trapezoid_params.update(trapezoid)
        trapezoid_params["length"] = trapezoid_params.get(
            "length", half_length - trapezoid_params["x"])
        self._initialize_feature(trapezoid_params)

        ref_circle = ref_circle or {}
        ref_circle_params = {
            "class": rf.RinkCircle,
            "y": -half_width,
            "thickness": line_thickness,
            "radius": 10,
            "color": line_color,
            "zorder": line_zorder,
        }
        ref_circle_params.update(ref_circle)
        self._initialize_feature(ref_circle_params)

        center_dot = center_dot or {}
        center_dot_params = {
            "class": rf.RinkCircle,
            "radius": red_line_params["length"] / 2,
            "color": "blue",
            "zorder": 11,
        }
        center_dot_params.update(center_dot)
        self._initialize_feature(center_dot_params)

        center_circle = center_circle or {}
        center_circle_params = {
            "class": rf.RinkCircle,
            "thickness": line_thickness,
            "radius": 15,
            "color": "blue",
            "zorder": line_zorder,
        }
        center_circle_params.update(center_circle)
        self._initialize_feature(center_circle_params)

        faceoff_circle = faceoff_circle or {}
        faceoff_circle_params = {
            "class": rf.FaceoffCircle,
            # 20' from front edge of goal line
            "x": goal_line_params["x"] - goal_line_params["length"] / 2 - 20,
            "y": 22,    # 44' between faceoff dots
            "length": 67 / 12,    # 5'7" between inside edges of hashmarks
            "width": 2,    # hashmarks are 2' long
            "thickness": line_thickness,
            "radius": center_circle_params["radius"],
            "resolution": 5000,    # increase resolution to keep lines straight
            "reflect_x": True,
            "reflect_y": True,
            "color": line_color,
            "zorder": line_zorder,
        }
        faceoff_circle_params.update(faceoff_circle)
        self._initialize_feature(faceoff_circle_params)

        ozone_dot_x = np.ravel(faceoff_circle_params["x"])
        dot_y = np.ravel(faceoff_circle_params["y"])

        faceoff_dot = faceoff_dot or {}
        faceoff_dot_params = {
            "class": rf.RinkCircle,
            # ozone faceoff circles and 5' from the blue line
            "x": [*ozone_dot_x, self._half_nzone_length - 5],
            "y": dot_y,
            "length": 16 / 12,    # edge to edge of inner shape
            "thickness": 1 / 12,
            "radius": 1,
            "reflect_x": True,
            "reflect_y": True,
            "color": "red",
            "zorder": 5,
        }

        # split dot into two shapes, one for the outer circle and one for the inner shape
        faceoff_dot_params.update(faceoff_dot)
        inner_dot_params = dict(faceoff_dot_params)
        inner_dot_params["class"] = rf.InnerDot
        self._initialize_feature(faceoff_dot_params)
        self._initialize_feature(inner_dot_params)

        faceoff_lines = faceoff_lines or {}
        faceoff_lines_params = {
            "class": rf.RinkL,
            "x": ozone_dot_x,
            "y": dot_y,
            "length": 4,
            "width": 3,
            "thickness": line_thickness,
            "reflect_x": True,
            "reflect_y": True,
            "color": line_color,
            "zorder": line_zorder,
        }
        faceoff_lines_params.update(faceoff_lines)

        try:
            iter(faceoff_lines_params["x"])
        except TypeError:
            faceoff_lines_params["x"] = [faceoff_lines_params["x"]]

        try:
            iter(faceoff_lines_params["y"])
        except TypeError:
            faceoff_lines_params["y"] = [faceoff_lines_params["y"]]

        # one L for each side of the dot
        for x_side in (1, -1):
            for y_side in (1, -1):
                current_line = dict(faceoff_lines_params)
                current_line["x"] = [x + x_dot_to_lines * x_side for x in current_line["x"]]
                current_line["y"] = [y + y_dot_to_lines * y_side for y in current_line["y"]]

                # change shape by using negative length and/or width
                current_line["length"] = current_line["length"] * x_side
                current_line["width"] = current_line["width"] * y_side

                self._initialize_feature(current_line)

        crease = crease or {}
        crease_params = {
            "class": rf.Crease,
            "x": goal_line_params["x"] - goal_line_params["length"] / 2,
            "length": 4.5,    # 4'6" rectangular section
            "width": 8,    # 8' from outside edge to outside edge
            "radius": 1.5,    # total length 6'
            "reflect_x": True,
            "reflect_y": True,
            "color": "lightblue",
        }
        crease_params.update(crease)

        crease_outline = crease_outline or {}
        crease_outline_params = dict(crease_params)
        crease_outline_params["thickness"] = line_thickness
        crease_outline_params["color"] = line_color
        crease_outline_params["zorder"] = line_zorder
        crease_outline_params.update(crease_outline)

        self._initialize_feature(crease_params)
        self._initialize_feature(crease_outline_params)

        crossbar = crossbar or {}
        crossbar_params = {
            "class": rf.Crossbar,
            "x": goal_line_params["x"] - goal_line_params["length"] / 2,
            # posts are 2 3/8" wide => half = 19/16"
            "radius": 19 / 16 / 12,
            "reflect_x": True,
            "color": "red",
            "zorder": 6,
        }
        crossbar_params.update(crossbar)
        crossbar_params["width"] = crossbar_params.get(
            "width", 6 + crossbar_params["radius"])
        self._initialize_feature(crossbar_params)

        net = net or {}
        net_params = {
            "class": rf.Net,
            "x": crossbar_params["x"] + crossbar_params["radius"] * 2,
            "length": 40 / 12,    # 40" deep
            "width": crossbar_params["width"] + crossbar_params["radius"],
            "thickness": 88 / 12,    # max width
            "radius": 20 / 12,
            "reflect_x": True,
            "color": "grey",
            "zorder": 5,
        }
        net_params.update(net)
        self._initialize_feature(net_params)

        for added_feature in added_features.values():
            self._initialize_feature(added_feature)


class NHLRink(Rink):
    """ Version of Rink class based off a typical NHL ice surface.

    Includes an additional feature "crease_notch" for the little notches inside the crease.

    See Rink for full documentation.
    """

    def __init__(self, **kwargs):
        crease = kwargs.get("crease", {})
        line_thickness = kwargs.get("line_thickness", 1 / 6)
        half_length = kwargs.get("boards", {}).get("length", 200) / 2
        half_goal_line_thickness = kwargs.get("goal_line", {}).get("length", line_thickness) / 2
        goal_line_x = kwargs.get("goal_line", {}).get(
            "x", half_length - 11 - half_goal_line_thickness)
        crease_thickness = crease.get("thickness", line_thickness)
        notch_width = 5 / 12

        nhl_updates = {
            "crease_notch": {
                "class": rf.RinkRectangle,
                "x": goal_line_x - 4 - crease_thickness / 2,
                "y": ((crease.get("width", 8) - notch_width) / 2
                      - crease_thickness),
                "length": crease_thickness,
                "width": notch_width,
                "reflect_x": crease.get("reflect_x", True),
                "reflect_y": crease.get("reflect_y", True),
                "color": kwargs.get("line_color", "red"),
                "zorder": kwargs.get("line_zorder", 5),
                "visible": crease.get("visible", True),
            },
        }

        kwargs["crease_notch"] = {**nhl_updates["crease_notch"],
                                  **kwargs.get("crease_notch", {})}

        super().__init__(**kwargs)


class NWHLRink(NHLRink):
    """ Version of Rink class based off of the NWHL rink in the 2021 playoffs.

    Includes additional features of "logo" for the logo at center ice and
    "crease_notch" for the little notches inside the crease.

    See Rink for full documentation.
    """

    def __init__(self, **kwargs):
        half_width = kwargs.get("boards", {}).get("width", 85) / 2
        center_radius = kwargs.get("center_circle", {}).get("radius", 15)

        nwhl_updates = {
            "nzone": {"color": "#B266FF", "visible": True},
            "ref_circle": {"y": half_width},
            "center_circle": {"thickness": 2, "color": "#003366", "zorder": 12},
            "center_dot": {"visible": False},
            "logo": {"class": rf.CircularImage,
                     "path": "https://raw.githubusercontent.com/the-bucketless/hockey_rink/master/images/nwhl_logo.png",
                     "radius": center_radius, "is_constrained": False, "zorder": 11}
        }

        for k, v in nwhl_updates.items():
            kwargs[k] = {**v, **kwargs.get(k, {})}

        super().__init__(**kwargs)


class IIHFRink(Rink):
    """ Version of Rink class with dimensions based off of IIHF regulations.

    Includes an additional feature "crease_notch" for the little notches inside the crease.

    See Rink for full documentation.
    """

    def __init__(self, **kwargs):
        iihf_updates = {
            "boards": {"length": 197, "width": 98.4},
            "nzone": {"length": 47},
            "trapezoid": {"visible": False},
            "crease": {"length": 0, "width": 11.6, "radius": 5.9},
        }

        line_thickness = kwargs.get("line_thickness", 1 / 6)
        half_goal_line_thickness = kwargs.get("goal_line", {}).get("length", line_thickness) / 2
        boards_length = kwargs.get("boards", {}).get("length", 197)
        iihf_updates["goal_line"] = {"x": boards_length / 2 - 13.1 + half_goal_line_thickness}
        goal_line_x = (kwargs.get("goal_line", iihf_updates["goal_line"])["x"]
                       - half_goal_line_thickness)

        iihf_updates["faceoff_circle"] = {"x": goal_line_x - 22}

        crease = kwargs.get("crease", {})
        crease_thickness = crease.get("thickness", line_thickness)
        notch_size = 5 / 12

        iihf_updates["crease_notch"] = {
            "class": rf.RinkL,
            "x": goal_line_x - 4,
            "y": 4,
            "length": notch_size,
            "width": -notch_size,
            "thickness": crease_thickness,
            "reflect_x": crease.get("reflect_x", True),
            "reflect_y": crease.get("reflect_y", True),
            "color": kwargs.get("line_color", "red"),
            "zorder": kwargs.get("line_zorder", 5),
            "visible": crease.get("visible", True),
        }

        for k, v in iihf_updates.items():
            kwargs[k] = {**v, **kwargs.get(k, {})}

        super().__init__(**kwargs)



""" Module for adding plotting functions to BaseRink.

Not intended for direct use, only as a parent class.
"""


from functools import wraps
from hockey_rink._base_rink import BaseRink
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import binned_statistic_2d


class BaseRinkPlot(BaseRink):
    """ Class extending BaseRink to include plotting methods. """

    def _validate_values(plot_function):
        """ Decorator to restrict x, y, and values to intended coordinates.

        Expects x and y to have already been converted for plotting.
        """

        @wraps(plot_function)
        def wrapper(self, x, y, *, values=None,
                    plot_range=None, plot_xlim=None, plot_ylim=None, **kwargs):

            C = kwargs.pop("C", None)
            values = C if values is None else values

            values = self.copy_(values)

            # x and y will already have been symmetrized
            if values is None:
                values = np.ones(x.shape)
            else:
                values = np.ravel(values)

                if kwargs.get("symmetrize", False):
                    values = np.concatenate((values, values))

            if len(x) != len(y) or len(x) != len(values):
                raise Exception("x, y, and values must have equal length.")

            mask = False

            if plot_range is None and plot_xlim is None and plot_ylim is None:
                plot_xlim, plot_ylim = self._get_limits("full")
            else:
                plot_xlim, plot_ylim = self._get_limits(plot_range,
                                                        self.copy_(plot_xlim),
                                                        self.copy_(plot_ylim))

                mask = ((x < plot_xlim[0]) | (x > plot_xlim[1])
                        | (y < plot_ylim[0]) | (y > plot_ylim[1]))

            if kwargs.get("is_constrained", True):
                values = self._outside_rink_to_nan(x, y, values)

            mask = mask | np.isnan(x) | np.isnan(y) | np.isnan(values)

            x = x[~mask]
            y = y[~mask]
            values = values[~mask]

            return plot_function(self, x, y, values=values,
                                 plot_range=plot_range, plot_xlim=plot_xlim, plot_ylim=plot_ylim,
                                 **kwargs)

        return wrapper

    def _validate_plot(plot_function):
        """ Decorator to ensure all plotting parameters are of the correct form. """

        @wraps(plot_function)
        def wrapper(self, *args, **kwargs):
            if "ax" not in kwargs:
                kwargs["ax"] = plt.gca()

            args = list(args)
            for coord in ("x", "y", "x1", "y1", "x2", "y2"):
                if coord in kwargs:
                    args.append(kwargs.pop(coord))

            for i in range(len(args)):
                args[i] = self.copy_(args[i])
                args[i] = np.ravel(args[i])

                is_y = i % 2

                if kwargs.get("symmetrize", False):
                    args[i] = np.concatenate((args[i], args[i] * (-1 if is_y else 1)))

                args[i] = args[i] - (self.y_shift if is_y else self.x_shift)

            kwargs["transform"] = self._get_transform(kwargs["ax"])

            args = tuple(args)

            return plot_function(self, *args, **kwargs)

        return wrapper

    def _constrain_plot(self, plot_features, ax, transform):
        """ Constrain the features of a plot to only display inside the boards. """

        try:
            iter(plot_features)
        except TypeError:
            plot_features = [plot_features]

        constraint = self._add_boards_constraint(ax, transform)

        for plot_feature in plot_features:
            plot_feature.set_clip_path(constraint)

    def _outside_rink_to_nan(self, x, y, values):
        """ Set values of coordinates outside the boundaries of the rink to nan. """

        x = np.abs(x).astype("float32")
        y = np.abs(y).astype("float32")

        values = values.astype("float32")

        center_x = self._boards_constraint.length / 2 - self._boards_constraint.radius
        center_y = self._boards_constraint.width / 2 - self._boards_constraint.radius

        mask = ((x > center_x) & (y > center_y)
                & ((center_x - x) ** 2 + (center_y - y) ** 2 > self._boards_constraint.radius ** 2))
        values[mask] = np.nan

        return values

    @staticmethod
    def binned_stat_2d(x, y, values, statistic="sum", xlim=None, ylim=None, binsize=1, bins=None):
        """ Use scipy to compute a bidimensional binned statistic.

        Parameters:
            x: array_like

            y: array_like

            values: array_like

            statistic: string or callable; default: "sum"
                The statistic to compute via scipy.  The following are available:
                    "mean": the mean of values for points within each bin.
                    "std": the standard deviation within each bin.
                    "median": the median of values for points within each bin.
                    "count": the number of points within each bin.
                    "sum": the sum of values for points within each bin.
                    "min": the minimum of values for points within each bin.
                    "max": the maximum of values for points within each bin.
                    function: a user-defined function which takes a 1D array of values and outputs a
                        single numerical statistic.

            xlim: (float, float); optional
                The lower and upper bounds of the x-coordinates to include.

                Only used if bins is None.

            ylim: (float, float); optional
                The lower and upper bounds of the y-coordinates to include.

                Only used if bins is None.

            binsize: float or (float, float); default: 1
                The size of the bins for a given portion of the rink.
                    float: the size of the bins for the two dimensions.
                    (float, float): the size of the bins in each dimension.

                Only used if bins is None.

            bins: int or (int, int) or array_like or (array, array); optional
                The bin specification:
                    int: the number of bins for the two dimensions.
                    (int, int): the number of bins in each dimension.
                    array_like: the bin edges for the two dimensions.
                    (array, array): the bin edges in each dimension.

        Returns:
            stat: (nx, ny) ndarray
                The values of the selected statistic in each two-dimensional bin.
            x_edge: (nx + 1) ndarray
                The bin edges along the first dimension.
            y_edge: (ny + 1) ndarray
                The bin edges along the second dimension.
        """

        if bins is None:
            try:
                iter(binsize)
            except TypeError:
                binsize = (binsize, binsize)

            # coordinates set as center of each bin
            x_edge = np.arange(xlim[0] - binsize[0] / 2, xlim[1] + binsize[0], binsize[0])
            y_edge = np.arange(ylim[0] - binsize[1] / 2, ylim[1] + binsize[1], binsize[1])

            bins = [x_edge, y_edge]

        stat, x_edge, y_edge, _ = binned_statistic_2d(x, y, values, statistic=statistic, bins=bins)

        stat = stat.T

        return stat, x_edge, y_edge

    @_validate_plot
    def plot(self, x, y, *, is_constrained=True, zorder=20, ax=None, **kwargs):
        """ Wrapper for matplotlib plot function.

        Will plot to areas out of view when full ice surface is not displayed.

        All parameters other than x and y require keywords.
            ie) plot(x, y, False) won't work, needs to be plot(x, y, is_constrained=False)

        Parameters:
            x: array_like

            y: array_like

            is_constrained: bool; default: True
                Indicates whether or not the plot is constrained to remain inside the boards.

            zorder: float; default: 20
                Determines which rink features the plot will draw over.

            ax: matplotlib Axes; optional
                Axes in which to draw the plot.  If not provided, will use the currently active Axes.

            **kwargs: Any other matplotlib plot properties; optional

        Returns:
            list of matplotlib Line2D
        """

        img = ax.plot(x, y, zorder=zorder, **kwargs)

        if is_constrained:
            self._constrain_plot(img, ax, kwargs["transform"])

        return img

    @_validate_plot
    def scatter(self, x, y, *, is_constrained=True, symmetrize=False,
                zorder=20, ax=None, **kwargs):
        """ Wrapper for matplotlib scatter function.

        Will plot to areas out of view when full ice surface is not displayed.

        All parameters other than x and y require keywords.
            ie) scatter(x, y, False) won't work, needs to be scatter(x, y, is_constrained=False)

        Parameters:
            x: array_like

            y: array_like

            symmetrize: bool; default: False
                Indicates whether or not to reflect the coordinates across the y-axis.

            is_constrained: bool; default: True
                Indicates whether or not the plot is constrained to remain inside the boards.

            zorder: float; default: 20
                Determines which rink features the plot will draw over.

            ax: matplotlib Axes; optional
                Axes in which to draw the plot.  If not provided, will use the currently active Axes.

            **kwargs: Any other matplotlib scatter properties; optional

        Returns:
            matplotlib PathCollection
        """

        img = ax.scatter(x, y, zorder=zorder, **kwargs)

        if is_constrained:
            self._constrain_plot(img, ax, kwargs["transform"])

        return img

    @_validate_plot
    def arrow(self, x1, y1, x2, y2, *, is_constrained=True,
              length_includes_head=True, head_width=1,
              zorder=20, ax=None, **kwargs):
        """ Wrapper for matplotlib arrow function.

        Will plot to areas out of view when full ice surface is not displayed.

        All parameters other than x1, y1, x2, and y2 require keywords.
            ie) arrow(x1, y1, x2, y2, False) won't work, needs to be arrow(x1, y1, x2, y2, is_constrained=False)

        Parameters:
            x1: array_like
                The starting x-coordinates of the arrows.

            y1: array_like
                The starting y-coordinates of the arrows.

            x2: array_like
                The ending x-coordinates of the arrows.

            y2: array_like
                The ending y-coordinates of the arrows.

            is_constrained: bool; default: True
                Indicates whether or not the plot is constrained to remain inside the boards.

            length_includes_head: bool; default: True
                Indicates if head of the arrow is to be included in calculating the length.

            head_width: float or None; default: 1
                Total width of the full arrow head.

            zorder: float; default: 20
                Determines which rink features the plot will draw over.

            ax: matplotlib Axes; optional
                Axes in which to draw the plot.  If not provided, will use the currently active Axes.

            **kwargs: Any other matplotlib arrow properties; optional

        Returns:
            list of matplotlib FancyArrow
        """

        dx = x2 - x1
        dy = y2 - y1

        arrows = []
        for i in range(len(x1)):
            arrows.append(ax.arrow(x1[i], y1[i], dx[i], dy[i],
                                   zorder=zorder, head_width=head_width,
                                   length_includes_head=length_includes_head, **kwargs))

        if is_constrained:
            self._constrain_plot(arrows, ax, kwargs["transform"])

        return arrows

    @_validate_plot
    @_validate_values
    def hexbin(self, x, y, *, values=None, is_constrained=True, symmetrize=False,
               plot_range=None, plot_xlim=None, plot_ylim=None,
               gridsize=None, binsize=1, zorder=2, ax=None, **kwargs):
        """ Wrapper for matplotlib hexbin function.

        Will plot to areas out of view when full ice surface is not displayed.
        Use plot_range, plot_xlim, and plot_ylim to restrict to the area within view.

        All parameters other than x and y require keywords.
            ie) hexbin(x, y, values) won't work, needs to be hexbin(x, y, values=values)

        Parameters:
            x: array_like

            y: array_like

            values: array_like; optional
                If None, values of 1 will be assigned to each x,y-coordinate provided.

            symmetrize: bool; default: False
                Indicates whether or not to reflect the coordinates and values across the y-axis.

            is_constrained: bool; default: True
                Indicates whether or not the plot is constrained to remain inside the boards.

                If plot ranges are used, also constrains coordinates included to remain inside the boards.

            plot_range: {"full", "half", "offense", "defense", "ozone", "dzone"}; optional
                Restricts the portion of the rink that can be plotted to.  Does so by removing values outside of
                the given range.

                Only affects x-coordinates and can be used in conjuction with ylim, but will be superceded by
                xlim if provided.

                "full": The entire length of the rink is displayed.
                "half" or "offense": The offensive half (largest x-coordinates) of the rink is displayed.
                "defense": The defensive half (smallest x-coordinates) of the rink is displayed.
                "ozone": The offensive zone (blue line to end boards) of the rink is displayed.
                "dzone": The defensive zone (end boards to blue line) of the rink is displayed.

                If no values are given for plot_range, plot_xlim, or plot_ylim, will use all coordinates provided.

            plot_xlim: float or (float, float); optional
                The range of x-coordinates to include in the plot.
                    float: the lower bound of the x-coordinates.  The upper bound will be the boards.
                    (float, float): The lower and upper bounds of the x-coordinates.

                If no values are given for plot_range, plot_xlim, or plot_ylim, will use all coordinates provided.

            plot_ylim: float or (float, float); optional
                The range of y-coordinates to include in the plot.
                    float: the lower bound of the y-coordinates.  The upper bound will be the boards.
                    (float, float): The lower and upper bounds of the y-coordinates.

                If no values are given for plot_range, plot_xlim, or plot_ylim, will use all coordinates provided.

            gridsize: int or (int, int); optional
                The grid specification:
                    int: the number of hexagons in the x-direction.  The number of hexagons in the y-direction is
                        chosen such that hexagons are approximately regular.
                    (int, int): the number of hexagons in both directions.

            binsize: float or (float, float); default: 1
                The size of the bins for a given portion of the rink.
                    float: the size of the bins for the two dimensions.
                    (float, float): the size of the bins in each dimension.

            zorder: float; default: 2
                Determines which rink features the plot will draw over.

            ax: matplotlib Axes; optional
                Axes in which to draw the plot.  If not provided, will use the currently active Axes.

            **kwargs: Any other matplotlib hexbin properties; optional

        Returns:
            matplotlib PolyCollection
        """

        try:
            iter(binsize)
        except:
            binsize = (binsize, binsize)

        default_gridsize = (int((plot_xlim[1] - plot_xlim[0]) / binsize[0]),
                            int((plot_ylim[1] - plot_ylim[0]) / binsize[1]))
        gridsize = gridsize or default_gridsize

        img = ax.hexbin(x, y, C=values, gridsize=gridsize, zorder=zorder, **kwargs)

        if is_constrained:
            self._constrain_plot(img, ax, kwargs["transform"])

        return img

    @_validate_plot
    @_validate_values
    def heatmap(self, x, y, *, values=None, symmetrize=False, is_constrained=True,
                plot_range=None, plot_xlim=None, plot_ylim=None,
                statistic="sum", binsize=1, bins=None,
                zorder=2, ax=None, **kwargs):
        """ Wrapper for matplotlib pcolormesh function.

        Will plot to areas out of view when full ice surface is not displayed.
        Use plot_range, plot_xlim, and plot_ylim to restrict to the area within view.

        All parameters other than x and y require keywords.
            ie) heatmap(x, y, values) won't work, needs to be heatmap(x, y, values=values)

        Parameters:
            x: array_like

            y: array_like

            values: array_like; optional
                If None, values of 1 will be assigned to each x,y-coordinate provided.

            symmetrize: bool; default: False
                Indicates whether or not to reflect the coordinates and values across the y-axis.

            is_constrained: bool; default: True
                Indicates whether or not the plot is constrained to remain inside the boards.

                If plot ranges are used, also constrains coordinates included to remain inside the boards.

            plot_range: {"full", "half", "offense", "defense", "ozone", "dzone"}; optional
                Restricts the portion of the rink that can be plotted to.  Does so by removing values outside of
                the given range.

                Only affects x-coordinates and can be used in conjuction with ylim, but will be superceded by
                xlim if provided.

                "full": The entire length of the rink is displayed.
                "half" or "offense": The offensive half (largest x-coordinates) of the rink is displayed.
                "defense": The defensive half (smallest x-coordinates) of the rink is displayed.
                "ozone": The offensive zone (blue line to end boards) of the rink is displayed.
                "dzone": The defensive zone (end boards to blue line) of the rink is displayed.

                If no values are given for plot_range, plot_xlim, or plot_ylim, will use all coordinates provided.

            plot_xlim: float or (float, float); optional
                The range of x-coordinates to include in the plot.
                    float: the lower bound of the x-coordinates.  The upper bound will be the boards.
                    (float, float): The lower and upper bounds of the x-coordinates.

                If no values are given for plot_range, plot_xlim, or plot_ylim, will use all coordinates provided.

            plot_ylim: float or (float, float); optional
                The range of y-coordinates to include in the plot.
                    float: the lower bound of the y-coordinates.  The upper bound will be the boards.
                    (float, float): The lower and upper bounds of the y-coordinates.

                If no values are given for plot_range, plot_xlim, or plot_ylim, will use all coordinates provided.

            statistic: string or callable; default: "sum"
                The statistic to compute via scipy.  The following are available:
                    "mean": the mean of values for points within each bin.
                    "std": the standard deviation within each bin.
                    "median": the median of values for points within each bin.
                    "count": the number of points within each bin.
                    "sum": the sum of values for points within each bin.
                    "min": the minimum of values for points within each bin.
                    "max": the maximum of values for points within each bin.
                    function: a user-defined function which takes a 1D array of values and outputs a
                        single numberical statistic.

            binsize: float or (float, float); default: 1
                The size of the bins for a given portion of the rink.
                    float: the size of the bins for the two dimensions.
                    (float, float): the size of the bins in each dimension.

                Only used if bins is None.

            bins: int or (int, int) or array_like or (array, array); optional
                The bin specification:
                    int: the number of bins for the two dimensions.
                    (int, int): the number of bins in each dimension.
                    array_like: the bin edges for the two dimensions.
                    (array, array): the bin edges in each dimension.

            zorder: float; default: 2
                Determines which rink features the plot will draw over.

            ax: matplotlib Axes; optional
                Axes in which to draw the plot.  If not provided, will use the currently active Axes.

            **kwargs: Any other matplotlib pcolormesh properties; optional

        Returns:
            matplotlib QuadMesh
        """

        stat, x_edge, y_edge = self.binned_stat_2d(x, y, values, statistic, plot_xlim, plot_ylim,
                                                   binsize, bins)

        img = ax.pcolormesh(x_edge, y_edge, stat, zorder=zorder, **kwargs)

        if is_constrained:
            self._constrain_plot(img, ax, kwargs["transform"])

        return img

    @_validate_plot
    @_validate_values
    def contour(self, x, y, *, values=None, fill=True, symmetrize=False, is_constrained=True,
                plot_range=None, plot_xlim=None, plot_ylim=None,
                statistic="sum", binsize=1, bins=None,
                zorder=2, ax=None, **kwargs):
        """ Wrapper for matplotlib contour and contourf functions.

        Will plot to areas out of view when full ice surface is not displayed.
        Use plot_range, plot_xlim, and plot_ylim to restrict to the area within view.

        All parameters other than x and y require keywords.
            ie) contour(x, y, values) won't work, needs to be contour(x, y, values=values)

        Parameters:
            x: array_like

            y: array_like

            values: array_like; optional
                If None, values of 1 will be assigned to each x,y-coordinate provided.

            fill: bool; default: True
                Indicates whether or not to fill in the contours.

            symmetrize: bool; default: False
                Indicates whether or not to reflect the coordinates and values across the y-axis.

            is_constrained: bool; default: True
                Indicates whether or not the plot is constrained to remain inside the boards.

                If plot ranges are used, also constrains coordinates included to remain inside the boards.

            plot_range: {"full", "half", "offense", "defense", "ozone", "dzone"}; optional
                Restricts the portion of the rink that can be plotted to.  Does so by removing values outside of
                the given range.

                Only affects x-coordinates and can be used in conjuction with ylim, but will be superceded by
                xlim if provided.

                "full": The entire length of the rink is displayed.
                "half" or "offense": The offensive half (largest x-coordinates) of the rink is displayed.
                "defense": The defensive half (smallest x-coordinates) of the rink is displayed.
                "ozone": The offensive zone (blue line to end boards) of the rink is displayed.
                "dzone": The defensive zone (end boards to blue line) of the rink is displayed.

                If no values are given for plot_range, plot_xlim, or plot_ylim, will use all coordinates provided.

            plot_xlim: float or (float, float); optional
                The range of x-coordinates to include in the plot.
                    float: the lower bound of the x-coordinates.  The upper bound will be the boards.
                    (float, float): The lower and upper bounds of the x-coordinates.

                If no values are given for plot_range, plot_xlim, or plot_ylim, will use all coordinates provided.

            plot_ylim: float or (float, float); optional
                The range of y-coordinates to include in the plot.
                    float: the lower bound of the y-coordinates.  The upper bound will be the boards.
                    (float, float): The lower and upper bounds of the y-coordinates.

                If no values are given for plot_range, plot_xlim, or plot_ylim, will use all coordinates provided.

            statistic: string or callable; default: "sum"
                The statistic to compute via scipy.  The following are available:
                    "mean": the mean of values for points within each bin.
                    "std": the standard deviation within each bin.
                    "median": the median of values for points within each bin.
                    "count": the number of points within each bin.
                    "sum": the sum of values for points within each bin.
                    "min": the minimum of values for points within each bin.
                    "max": the maximum of values for points within each bin.
                    function: a user-defined function which takes a 1D array of values and outputs a
                        single numberical statistic.

            binsize: float or (float, float); default: 1
                The size of the bins for a given portion of the rink.
                    float: the size of the bins for the two dimensions.
                    (float, float): the size of the bins in each dimension.

                Only used if bins is None.

            bins: int or (int, int) or array_like or (array, array); optional
                The bin specification:
                    int: the number of bins for the two dimensions.
                    (int, int): the number of bins in each dimension.
                    array_like: the bin edges for the two dimensions.
                    (array, array): the bin edges in each dimension.

            zorder: float; default: 2
                Determines which rink features the plot will draw over.

            ax: matplotlib Axes; optional
                Axes in which to draw the plot.  If not provided, will use the currently active Axes.

            **kwargs: Any other matplotlib contour properties; optional

        Returns:
            matplotlib QuadContourSet
        """

        stat, x_edge, y_edge = self.binned_stat_2d(x, y, values, statistic, plot_xlim, plot_ylim,
                                                   binsize, bins)

        x_centers = (x_edge[:-1] + x_edge[1:]) / 2
        y_centers = (y_edge[:-1] + y_edge[1:]) / 2

        if plot_xlim is not None:
            x_centers[-1] = max(x_centers[-1], plot_xlim[1])
        if plot_ylim is not None:
            y_centers[-1] = max(y_centers[-1], plot_ylim[1])

        if fill:
            img = ax.contourf(x_centers, y_centers, stat, zorder=zorder, **kwargs)
        else:
            img = ax.contour(x_centers, y_centers, stat, zorder=zorder, **kwargs)

        if is_constrained:
            self._constrain_plot(img.collections, ax, kwargs["transform"])

        return img

    """ Alias for contour. """
    contourf = contour



""" Abstract base class for drawing a hockey rink.

Not intended for direct use, only as a parent class.
"""


from abc import ABC
import hockey_rink.rink_feature as rf
import matplotlib.pyplot as plt
from matplotlib.transforms import Affine2D
import numpy as np


class BaseRink(ABC):
    """ Abstract base class for drawing rinks using matplotlib.

    Attributes:
        x_shift: float
            Amount x-coordinates are to be shifted.

            When viewing the rink horizontally, the coordinate of the center of the ice surface from left to right.
                eg) If using data with a coordinate system that goes from 0 to 200, x_shift should be 100.

            The actual coordinates won't be affected.  The purpose is to update the coordinates passed in to
            align with the drawing, not to alter the drawing to align with the coordinates.

        y_shift: float
            Amount y-coordinates are to be shifted.

            When viewing the rink horizontally, the coordinate of the center of the ice surface from bottom to top.
                eg) If using data with a coordinate system that goes from 0 to 85, y_shift should be 42.5.

            The actual coordinates won't be affected.  The purpose is to update the coordinates passed in to
            align with the drawing, not to alter the drawing to align with the coordinates.
    """

    def __init__(self):
        self.x_shift = 0
        self.y_shift = 0
        self._rotation = None
        self._feature_xlim = None
        self._feature_ylim = None
        self._features = []
        self._boards_constraint = None
        self._half_nzone_length = 0
        self._blue_line_thickness = 0

    def _initialize_feature(self, params):
        """ Initialize a feature of the rink at each coordinate it's required.

        Parameters:
            params: dict
                The class of the feature and all parameters required to instantiate its class.
        """

        feature_class = params.pop("class")

        xs = np.ravel(params.get("x", [0]))
        ys = np.ravel(params.get("y", [0]))

        x_reflections = [False, True] if params.pop("reflect_x", False) else [False]
        y_reflections = [False, True] if params.pop("reflect_y", False) else [False]

        for x in xs:
            for y in ys:
                for x_reflection in x_reflections:
                    for y_reflection in y_reflections:
                        feature_params = dict(params)

                        feature_params["x"] = x * (-1 if x_reflection else 1)
                        feature_params["y"] = y * (-1 if y_reflection else 1)
                        feature_params["is_reflected_x"] = x_reflection
                        feature_params["is_reflected_y"] = y_reflection

                        self._features.append(feature_class(**feature_params))

    def _add_boards_constraint(self, ax, transform=None):
        """ Add the boards constraint to the rink to avoid features extending beyond boards.

        Parameters:
            ax: matplotlib Axes
                Axes in which to add the constraint.

            transform: matplotlib Transform; optional
                Transform to apply to the constraint.

        Returns:
            matplotlib Polygon
        """

        transform = transform or ax.transData

        constraint = self._boards_constraint.get_polygon()
        constraint.set_transform(transform)
        ax.add_patch(constraint)

        return constraint

    def _get_limits(self, display_range="full", xlim=None, ylim=None, thickness=0):
        """ Return the xlim and ylim values corresponding to the parameters.

        Doesn't allow for coordinates extending beyond the extent of the boards.
        """

        xlim = self.copy_(xlim)
        ylim = self.copy_(ylim)

        # if boards are included in the limits, need to include their thickness
        half_length = self._boards_constraint.length / 2 + thickness

        if xlim is None:
            equivalencies = {
                "half": "offense",
                "offence": "offense",
                "defence": "defense",
            }
            display_range = display_range.lower() \
                .replace(" ", "")
            display_range = equivalencies.get(display_range, display_range)

            xlims = {
                "offense": (0, half_length),
                "defense": (-half_length, 0),
                "ozone": (self._half_nzone_length, half_length),
                "nzone": (-self._half_nzone_length - self._blue_line_thickness,
                          self._half_nzone_length + self._blue_line_thickness),
                "dzone": (-half_length, -self._half_nzone_length),
            }

            xlim = xlims.get(display_range, (-half_length, half_length))
        else:
            try:
                xlim = (xlim[0] - self.x_shift, xlim[1] - self.x_shift)
            except TypeError:
                xlim -= self.x_shift

                if xlim >= half_length:
                    xlim = -half_length

                xlim = (xlim, half_length)

        half_width = self._boards_constraint.width / 2 + thickness
        if ylim is None:
            ylim = (-half_width, half_width)
        else:
            try:
                ylim = (ylim[0] - self.y_shift, ylim[1] - self.y_shift)
            except TypeError:
                ylim -= self.y_shift

                if ylim >= half_width:
                    ylim = -half_width

                ylim = (ylim, half_width)

        # always have the smaller coordinate first
        if xlim[0] > xlim[1]:
            xlim = (xlim[1], xlim[0])
        if ylim[0] > ylim[1]:
            ylim = (ylim[1], ylim[0])

        # disallow coordinates from extending beyond the rink
        xlim = (max(xlim[0], -half_length), min(xlim[1], half_length))
        ylim = (max(ylim[0], -half_width), min(ylim[1], half_width))

        return xlim, ylim

    def _get_transform(self, ax):
        """ Return the matplotlib Transform to apply to features of the rink. """

        return self._rotation + ax.transData

    @staticmethod
    def copy_(param):
        """ Return a copy of a parameter where possible. """

        try:
            param = param.copy()
        except AttributeError:
            pass

        return param

    def convert_xy(self, x, y):
        """ Convert x,y-coordinates to the scale used for the rink. """

        x = self.copy_(x)
        y = self.copy_(y)

        x = np.ravel(x) - self.x_shift
        y = np.ravel(y) - self.y_shift

        if self._rotation:
            xy = self._rotation.transform(tuple(zip(x, y)))
            return xy[:, 0], xy[:, 1]
        else:
            return x, y

    def draw(self, ax=None, display_range="full", xlim=None, ylim=None, rotation=None):
        """ Draw the rink.

        Parameters:
            ax: matplotlib Axes; optional
                Axes in which to draw the rink.  If not provided, the currently-active Axes is used.

            display_range: {"full", "half", "offense", "defense", "ozone", "dzone"}; default: "full"
                The portion of the rink to display.  The entire rink is drawn regardless, display_range only
                affects what is shown.

                Only affects x-coordinates and can be used in conjuction with ylim, but will be superceded by
                xlim if provided.

                If a rotation other than those resulting in the rink being drawn horizontal/vertical, coordinates
                outside of the display range may be included.

                Features not constrained within the boards (not including the boards) will only be displayed if
                display_range is "full".

                "full": The entire length of the rink is displayed.
                "half" or "offense": The offensive half (largest x-coordinates) of the rink is displayed.
                "defense": The defensive half (smallest x-coordinates) of the rink is displayed.
                "ozone": The offensive zone (blue line to end boards) of the rink is displayed.
                "dzone": The defensive zone (end boards to blue line) of the rink is displayed.

                If no acceptable values for display_range, xlim, and ylim are provided, will display the full rink.

            xlim: float or (float, float); optional
                If a single float, the lower bound of the x-coordinates to display.  The upper bound will be the
                end of the boards.

                If a tuple, the lower and upper bounds of the x-coordinates to display.

            ylim: float or (float, float); optional
                If a single float, the lower bound of the y-coordinates to display.  The upper bound will be the
                end of the boards.

                If a tuple, the lower and upper bounds of the y-coordinates to display.

            rotation: float; optional
                Degrees to rotate the rink when drawing.

                If used, sets the class attribute.

                0 corresponds to a horizontal rink with largest x and y-coordinates in the top right quadrant.
                90 will rotate so that the largest coordinates are in the top left quadrant.

        Returns:
            matplotlib Axes
        """

        if rotation is not None:
            self._rotation = Affine2D().rotate_deg(rotation)

        if ax is None:
            ax = plt.gca()

        ax.set_aspect("equal")
        ax.axis("off")

        transform = self._get_transform(ax)

        constraint = self._add_boards_constraint(ax, transform)

        for feature in self._features:
            drawn_feature = feature.draw(ax, transform)

            if feature.is_constrained:
                try:
                    drawn_feature.set_clip_path(constraint)
                except AttributeError:
                    pass
            else:
                # need to track outer bounds of unconstrained features to properly set xlim and ylim
                try:
                    visible = feature.visible
                except AttributeError:
                    visible = True

                if visible and not isinstance(feature, rf.Boards):
                    try:
                        feature_x, feature_y = feature.get_polygon_xy()

                        if self._feature_xlim is None:
                            self._feature_xlim = [np.min(feature_x), np.max(feature_x)]
                        else:
                            print(self._feature_xlim)
                            self._feature_xlim = [min(self._feature_xlim[0], np.min(feature_x)),
                                                  max(self._feature_xlim[1], np.max(feature_x))]

                        if self._feature_ylim is None:
                            self._feature_ylim = [np.min(feature_y), np.max(feature_y)]
                        else:
                            self._feature_ylim = [min(self._feature_ylim[0], np.min(feature_y)),
                                                  max(self._feature_ylim[1], np.max(feature_y))]
                    except TypeError:
                        pass

        ax = self.set_display_range(ax, display_range, xlim, ylim)

        return ax

    def set_display_range(self, ax=None, display_range="full", xlim=None, ylim=None):
        """ Set the xlim and ylim for the matplotlib Axes.

        Parameters:
            ax: matplotlib Axes; optional
                Axes in which to set xlim and ylim.

            display_range: {"full", "half", "offense", "defense", "ozone", "dzone"}; default: "full"
                The portion of the rink to display.  The entire rink is drawn regardless, display_range only
                affects what is shown.

                Only affects x-coordinates and can be used in conjuction with ylim, but will be superceded by
                xlim if provided.

                If a rotation other than those resulting in the rink being drawn horizontal/vertical, coordinates
                outside of the display range may be included.

                Features not constrained within the boards (not including the boards) will only be displayed if
                display_range is "full".

                "full": The entire length of the rink is displayed.
                "half" or "offense": The offensive half (largest x-coordinates) of the rink is displayed.
                "defense": The defensive half (smallest x-coordinates) of the rink is displayed.
                "ozone": The offensive zone (blue line to end boards) of the rink is displayed.
                "dzone": The defensive zone (end boards to blue line) of the rink is displayed.

                If no acceptable values for display_range, xlim, and ylim are provided, will display the full rink.

            xlim: float or (float, float); optional
                If a single float, the lower bound of the x-coordinates to display.  The upper bound will be the
                end of the boards.

                If a tuple, the lower and upper bounds of the x-coordinates to display.

            ylim: float or (float, float); optional
                If a single float, the lower bound of the y-coordinates to display.  The upper bound will be the
                end of the boards.

                If a tuple, the lower and upper bounds of the y-coordinates to display.
        """

        if ax is None:
            ax = plt.gca()

        x, y = self._boards_constraint.get_polygon_xy()

        xlim, ylim = self._get_limits(display_range, xlim, ylim,
                                      self._boards_constraint.thickness)

        mask = (x >= xlim[0]) & (x <= xlim[1]) & (y >= ylim[0]) & (y <= ylim[1])
        x = np.concatenate((x[mask], xlim, xlim))
        y = np.concatenate((y[mask], ylim, ylim[::-1]))

        if display_range == "full":
            if self._feature_xlim is not None:
                x = np.concatenate((x, self._feature_xlim))
            if self._feature_ylim is not None:
                y = np.concatenate((y, self._feature_ylim))

        # need to shift so convert_xy can reverse shift
        xs, ys = self.convert_xy(x + self.x_shift, y + self.y_shift)

        ax.set_xlim(np.min(xs), np.max(xs))
        ax.set_ylim(np.min(ys), np.max(ys))

        return ax



from setuptools import find_packages, setup


with open("README.md") as readme_file:
    readme = readme_file.read()

setup(
    name="hockey_rink",
    version="0.1.1",
    description="A Python library for plotting hockey rinks with Matplotlib.",
    long_description_content_type="text/markdown",
    long_description=readme,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Matplotlib",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.6",
    ],
    url="https://github.com/the-bucketless/hockey_rink",
    author="The Bucketless",
    author_email="thebucketless@protonmail.com",
    license="GNU General Public License v3 (GPLv3)",
    packages=find_packages(),
    install_requires=["matplotlib", "numpy", "scipy"],
    zip_safe=False,
)



""" Module containing all provided rink features as well as the abstract base class.

Currently available features are:
    Boards
    BoardsConstraint
    RinkRectangle
    RinkCircle
    TrapezoidLine
    InnerDot
    FaceoffCircle
    RinkL
    Crease
    Crossbar
    Net
    CircularImage
"""


__all__ = ["RinkFeature", "Boards", "BoardsConstraint", "RinkRectangle",
           "RinkCircle", "TrapezoidLine", "InnerDot", "FaceoffCircle",
           "RinkL", "Crease", "Crossbar", "Net", "CircularImage"]


from abc import ABC, abstractmethod
import matplotlib.pyplot as plt
from matplotlib.transforms import Affine2D
import numpy as np


class RinkFeature(ABC):
    """ Abstract base class for features of a rink to draw.

    Attributes:
        x: float

        y: float

        length: float
            Typically, the size of the feature from left to right.

        width: float
            Typically, the size of the feature from bottom to top.

        thickness: float

        radius: float

        resolution: int
            The number of coordinates used in creating arcs.

        x_reflection: {-1, 1}
            If -1, reflects all x-coordinates.
            If 1, leaves x-coordinates as they are.

        y_reflection: {-1, 1}
            If -1, reflects all y-coordinates.
            If 1, leaves y-coordinates as they are.

        is_constrained: bool
            Indicates whether or not the feature is constrained to remain inside the boards.

        visible: bool

        polygon_kwargs: additional keyword arguments
            Any additional arguments to be passed to matplotlib's Polygon.
    """

    def __init__(self, x=0, y=0, length=0, width=0, thickness=0, radius=0,
                 resolution=500, is_reflected_x=False, is_reflected_y=False,
                 is_constrained=True, visible=True, **polygon_kwargs):
        """ Initialize attributes.

        Parameters:
            x: float; default: 0

            y: float; default: 0

            length: float; default: 0

            width: float; default: 0

            thickness: float; default: 0

            radius: float; default: 0

            resolution: int; default: 500

            is_reflected_x: bool; default: False
                Indicates if x-coordinates are to be reflected.

            y_reflection: bool; default: False
                Indicates if y-coordinates are to be reflected.

            is_constrained: bool; default: True

            visible: bool; default: True

            polygon_kwargs: additional keyword arguments; optional
        """

        self.x = x
        self.y = y
        self.length = length
        self.width = width
        self.thickness = thickness
        self.radius = radius
        self.resolution = resolution
        self.x_reflection = -1 if is_reflected_x else 1
        self.y_reflection = -1 if is_reflected_y else 1
        self.is_constrained = is_constrained
        self.visible = visible
        self.polygon_kwargs = polygon_kwargs

    @abstractmethod
    def _get_centered_xy(self):
        """ Abstract method to return the x and y-coordinates necessary to create a Polygon of the feature
        were it to be recorded at (0, 0).

        The actual coordinates used are left to get_polygon_xy.

        Returns:
            numpy ndarray, numpy ndarray
        """

        pass

    @staticmethod
    def arc_coords(center, width, height=None, thickness=0, theta1=0, theta2=360, resolution=500):
        """" Return the coordinates of an arc.

        Adapted from: https://stackoverflow.com/questions/30642391/how-to-draw-a-filled-arc-in-matplotlib

        Parameters:
            center: tuple
                The center point of the arc.

            width: float

            height: float; optional
                If not provided, height will be equal to width, creating a circular arc.

            thickness: float; default: 0
                If not 0, will arcs: one with the width and height provided and one with the width plus the
                thickness and the height plus the thickness.

            theta1: float; default: 0
                Degree at which to start the arc.

                0 corresponds to directly to the right of center, 90 to directly above, etc.

            theta2: float; default: 360
                Degree at which to end the arc.

                0 corresponds to directly to the right of center, 90 to directly above, etc.

            resolution: int; default: 500
                The number of coordinates used in creating the arc.

        Returns:
            numpy ndarray, numpy ndarray
        """

        height = width if height is None else height

        theta = np.linspace(np.radians(theta1), np.radians(theta2), resolution)
        x = width * np.cos(theta) + center[0]
        y = height * np.sin(theta) + center[1]

        if thickness > 0:
            theta = np.linspace(np.radians(theta2), np.radians(theta1), resolution)
            x = np.concatenate((x, (width + thickness) * np.cos(theta) + center[0]))
            y = np.concatenate((y, (height + thickness) * np.sin(theta) + center[1]))

        return x, y

    @staticmethod
    def tangent_point(center, radius, point):
        """ Return the coordinates on a circle that are tangent to a point outside the circle.

        Adapted from: https://stackoverflow.com/a/49987361

        Parameters:
            center: tuple
                The center point of the circle.

            radius: float

            point: tuple
                The coordinates of the point outside the circle.

        Returns:
            float, float
        """

        dx = point[0] - center[0]
        dy = point[1] - center[1]
        center_to_point = (dx ** 2 + dy ** 2) ** 0.5
        theta = np.arccos(radius / center_to_point)
        angle = np.arctan2(dy, dx) - theta * np.sign(point[1])

        return center[0] + radius * np.cos(angle), center[1] + radius * np.sin(angle)

    def get_polygon(self):
        """ Return the matplotlib Polygon representing the feature. """

        polygon_x, polygon_y = self.get_polygon_xy()

        return plt.Polygon(tuple(zip(polygon_x, polygon_y)), visible=self.visible, **self.polygon_kwargs)

    def get_polygon_xy(self):
        """ Return numpy ndarray, numpy ndarray of the x,y-coordinates necessary for creating the
        matplotlib Polygon representing the feature.
        """

        x, y = self._get_centered_xy()
        x = x * self.x_reflection + self.x
        y = y * self.y_reflection + self.y

        return x, y

    def draw(self, ax, transform=None):
        """ Draw the feature.

        Parameters:
            ax: matplotlib Axes
                Axes in which to draw the feature.

            transform: matplotlib Transform; optional
                Transform to apply to the feature.

        Returns:
            matplotlib Polygon
        """

        transform = transform or ax.transData

        patch = self.get_polygon()
        ax.add_patch(patch)
        patch.set_transform(transform)

        return patch


class Boards(RinkFeature):
    """ One quarter of a rink's boards.

    length and width refer to the full ice surface, not this specific quadrant.
    radius is for the arc in the corner boards.

    See RinkFeature for full documentation.
    """

    def _get_centered_xy(self):
        end_x = self.length / 2
        end_y = self.width / 2

        center_x = end_x - self.radius
        center_y = end_y - self.radius

        arc_x, arc_y = self.arc_coords((center_x, center_y), width=self.radius,
                                       thickness=self.thickness, theta1=90, theta2=0)

        inner_arc_x, outer_arc_x = np.reshape(arc_x, (2, -1))
        inner_arc_y, outer_arc_y = np.reshape(arc_y, (2, -1))

        board_x = np.concatenate((
            [0],    # inside center ice
            inner_arc_x,    # inside corner
            [end_x],    # inside end boards
            [end_x + self.thickness],    # outside end boards
            outer_arc_x,    # outside corner
            [0],    # outside center ice
        ))

        board_y = np.concatenate((
            [end_y],
            inner_arc_y,
            [0],
            [0],
            outer_arc_y,
            [end_y + self.thickness],
        ))

        return board_x, board_y


class BoardsConstraint(RinkFeature):
    """ The inside edge of the boards.  To be used to constrain other features to being inside the rink.

    Unlike boards, is used for the entire ice surface.

    radius is for the arc in the corner boards.

    See RinkFeature for full documentation.
    """

    def _get_centered_xy(self):
        end_x = self.length / 2
        end_y = self.width / 2

        center_x = end_x - self.radius
        center_y = end_y - self.radius

        arc_x, arc_y = self.arc_coords((center_x, center_y), width=self.radius,
                                       theta1=90, theta2=0, resolution=self.resolution)

        board_x = np.concatenate((
            arc_x,    # top right corner
            arc_x[::-1],    # bottom right corner
            -arc_x,    # bottom left corner
            -arc_x[::-1],    # top left corner
        ))

        board_y = np.concatenate((
            arc_y,
            -arc_y[::-1],
            -arc_y,
            arc_y[::-1],
        ))

        return board_x, board_y


class RinkRectangle(RinkFeature):
    """ A rectangle to be drawn on the rink (typically lines).

    x and y refer to the center of the rectangle.
    length is always left to right.
    width is always bottom to top.

    See RinkFeature for full documentation.
    """

    def _get_centered_xy(self):
        half_length = self.length / 2
        half_width = self.width / 2
        x = np.array([-half_length, half_length, half_length, -half_length])
        y = np.array([-half_width, -half_width, half_width, half_width])

        return x, y


class RinkCircle(RinkFeature):
    """ A circle to be drawn on the rink.

    x and y refer to the center of the circle.
    When including thickness, radius is distance to the outer edge, not the inner.

    See RinkFeature for full documentation.
    """

    def _get_centered_xy(self):
        return self.arc_coords((0, 0), width=self.radius - self.thickness,
                               thickness=self.thickness, resolution=self.resolution)


class TrapezoidLine(RinkFeature):
    """ One line of a trapezoid (typically for the goalie's restricted area).

    length is distance from left to right.
    width is distance from bottom to top.

    See RinkFeature for full documentation.
    """

    def _get_centered_xy(self):
        half_thickness = self.thickness / 2

        x = np.array([0, 0, self.length, self.length])
        y = np.array([-half_thickness, half_thickness,
                      self.width + half_thickness, self.width - half_thickness])

        return x, y


class InnerDot(RinkFeature):
    """ The shape inside a faceoff dot.

    x and y refer to the center of the dot.
    length is from the center to the edge of the opening.
    thickness is the size of the edge of the circle.

    See RinkFeature for full documentation.
    """

    def _get_centered_xy(self):
        radius = self.radius - self.thickness
        circle_x, circle_y = self.arc_coords((0, 0), width=radius,
                                             resolution=self.resolution)

        half_length = self.length / 2
        mask = (half_length >= circle_x) & (-half_length <= circle_x)

        return circle_x[mask], circle_y[mask]


class FaceoffCircle(RinkFeature):
    """ A circle including hashmarks.

    x and y refer to the center of the circle.
    length is distance between inside edges of the hashmarks.
    width is length of the hashmarks.
    radius is distance to the outer edge, not the inner.

    See RinkFeature for full documentation.
    """

    def _get_centered_xy(self):
        circle_x, circle_y = self.arc_coords((0, 0), width=self.radius - self.thickness,
                                             thickness=self.thickness, resolution=self.resolution)

        inner_circle_x, outer_circle_x = np.reshape(circle_x, (2, -1))
        inner_circle_y, outer_circle_y = np.reshape(circle_y, (2, -1))

        half_length = self.length / 2

        # end of the hashmark
        hashmark_y = (self.radius ** 2 - half_length ** 2) ** 0.5 + self.thickness + self.width
        hashmark = np.full(outer_circle_y.shape, hashmark_y) * np.sign(outer_circle_y)

        # replace the y coordinates on the outer edge of the circle with the end of the hashmark
        mask = ((np.abs(outer_circle_x) >= self.width)
                & (np.abs(outer_circle_x) <= self.width + self.thickness))
        outer_circle_y[mask] = hashmark[mask]

        x = np.concatenate((inner_circle_x, outer_circle_x))
        y = np.concatenate((inner_circle_y, outer_circle_y))

        return x, y


class RinkL(RinkFeature):
    """ L shapes to be drawn on the rink (typically for faceoff lines).

    x and y refer to the inside corner of the L.
    length is the | of the L.
    width is the _ of the L.

    See RinkFeature for full documentation.
    """

    def _get_centered_xy(self):
        # inside corner -> top of L -> outside corner -> right of L
        x = np.array(([0, 0, self.thickness * np.sign(self.length),
                       self.thickness * np.sign(self.length), self.length, self.length]))
        y = np.array(([0, self.width, self.width, self.thickness * np.sign(self.width),
                       self.thickness * np.sign(self.width), 0]))

        return x, y


class Crease(RinkFeature):
    """ A goaltender's crease.

    Used for both the crease (when thickness is 0) and the outline of the crease (when thickness isn't 0).

    Typically a rectangle with an arc at the end, but the rectangle can be removed by setting length to 0.

    x is goal line edge of crease.
    y is center of crease.
    length is the length of the rectangular portion of the crease.
    width is the width of the crease including the thickness of the outline.

    See RinkFeature for full documentation.
    """

    def _get_centered_xy(self):
        half_width = self.width / 2

        arc_x, arc_y = self.arc_coords(center=(self.length, 0),
                                       width=self.radius - self.thickness,
                                       height=half_width - self.thickness,
                                       thickness=self.thickness,
                                       theta1=90, theta2=0,
                                       resolution=self.resolution)

        if self.thickness == 0:    # crease
            x = np.concatenate(([0, 0], -arc_x))
            y = np.concatenate(([0, half_width], arc_y))
        else:    # crease outline
            x = np.concatenate(([0], -arc_x, [0]))
            y = np.concatenate(([half_width - self.thickness], arc_y, [half_width]))

        return x, y


class Crossbar(RinkFeature):
    """ The crossbar of a net.

    x is the front edge of the goal line.
    y is the center of the net.
    width doesn't include the radius of the bar.
    radius is used to create an arc at each end of the bar.

    See RinkFeature for full documentation.
    """

    def _get_centered_xy(self):
        half_width = self.width / 2

        # rounded end of the bar
        arc_x, arc_y = self.arc_coords((self.radius, 0), width=self.radius,
                                       theta2=180, resolution=self.resolution)

        x = np.concatenate((arc_x, arc_x[::-1]))
        y = np.concatenate((arc_y + half_width, -arc_y[::-1] - half_width))

        return x, y


class Net(RinkFeature):
    """ The netting of a hockey net.

    x is the back edge of the crossbar.
    y is the center of the net.
    length is the depth of the net.
    width is from the end of one post to the end of the other.
    thickness is the outer width (including the arc at the back).
    radius is for the arc at the back of the net.

    See RinkFeature for full documentation.
    """

    def _get_centered_xy(self):
        half_width = self.width / 2
        half_outer_width = max(half_width, self.thickness / 2)

        center_x = self.length - self.radius
        center_y = half_outer_width - self.radius

        arc_x, arc_y = self.arc_coords((center_x, center_y), width=self.radius,
                                       theta1=180, theta2=0, resolution=self.resolution)
        tangent_x, tangent_y = self.tangent_point((center_x, center_y), self.radius, (0, half_width))

        mask = (0 < arc_x) & (arc_x > tangent_x) & (arc_y > -half_outer_width)
        arc_x = arc_x[mask]
        arc_y = arc_y[mask]

        x = np.concatenate((arc_x, arc_x[::-1], [0, 0]))
        y = np.concatenate((arc_y, -arc_y[::-1], [-half_width, half_width]))

        return x, y


class CircularImage(RinkCircle):
    """ An image to be cropped inside a circle.

    Requires a path variable of the image file to be read.
    x and y refer to the center of the circle.
    If is_constrained is set to False, the image will no longer be circular.

    See RinkFeature for full documentation.
    """

    def __init__(self, path, rotation=0, is_constrained=False, **polygon_kwargs):
        self.path = path
        self.rotation = rotation
        super().__init__(is_constrained=is_constrained, **polygon_kwargs)

    def draw(self, ax, transform=None):
        """ Draw the feature.

        Parameters:
            ax: matplotlib Axes
                Axes in which to draw the feature.

            transform: matplotlib Transform; optional
                Transform to apply to the feature.

        Returns:
            matplotlib AxesImage
        """

        if not self.visible:
            return None

        transform = transform or ax.transData

        try:
            image = plt.imread(self.path)

            x = self.x * self.x_reflection
            y = self.y * self.y_reflection

            extent = [int(x - self.radius), int(x + self.radius),
                      int(y - self.radius), int(y + self.radius)]

            im = ax.imshow(image, extent=extent, **self.polygon_kwargs)
            im.set_transform(Affine2D().rotate_deg_around(self.x, self.y, self.rotation)
                             + transform)

            patch = plt.Circle((x, y), radius=self.radius, transform=transform)
            im.set_clip_path(patch)

            return im

        except Exception as e:
            print(e)

            return None
